@@ -1,22 +1,24 @@
 #include "../config.h"
 
-#include <string.h>
 #include <gtk/gtklabel.h>
 #include <gtk/gtk.h>
-#include <stdlib.h>
 #include <glibtop.h>
 #include <glibtop/cpu.h>
 #include <libgnomeui/libgnomeui.h>
-#include <math.h>
 #include <glade/glade.h>
 
+#include <stdlib.h>
+#include <string.h>
+#include <math.h>
+#include <sys/time.h>
+#include <time.h>
+
+
 #include <gnome.h>
 #include <libgnome/libgnome.h>
 #include <panel-applet.h>
 #include <panel-applet-gconf.h>
 
-#define AVERAGE_NUM_TIMES ((int) 32)
-
 typedef struct {
 	PanelApplet base;
 	GtkTooltips* tooltips;
@@ -27,15 +29,17 @@ typedef struct {
 	GtkWidget* prefs_dialog;
 
 	guint timeout_handler_id;
-	int ref_time_idx;
 	int time_idx;
 	int tooltip_update;
 	glibtop_cpu* cpu;
+	double timeout_change_time;
 
 	gboolean config_change;
 
 	gint flame_sensitivity;
-	gint update_interval;
+	gint cur_update_interval;
+	gint min_update_interval;
+	gint max_update_interval;
 	int decay;
 	int border;
 	int panel_border;
@@ -85,6 +89,9 @@ static unsigned int cpufire_flame[256] = {
   0xf9f6d4, 0xf9f6d4, 0xf9f6d4, 0xf9f6d4, 0xf9f6d4, 0xf9f6d4, 0xf9f6d4, 0xf9f6d4
 };
 
+static const double powersave_load_limit=0.2;
+static const int powersave_area_limit=10;  /* powersave for 100/x% of the area */
+
 static void start_procman_cb (BonoboUIComponent* uic, CPUFireApplet* applet);
 static void cpufire_applet_properties_dialog (BonoboUIComponent* uic, CPUFireApplet* applet);
 static void cpufire_applet_help_cb (BonoboUIComponent* uic, CPUFireApplet* applet);
@@ -124,32 +131,36 @@ cpufire_applet_get_type (void)
 static void
 cpufire_applet_load_properties (CPUFireApplet *applet)
 {
-	applet->update_interval = panel_applet_gconf_get_int (PANEL_APPLET (applet), "update_interval", NULL);
+	applet->min_update_interval = panel_applet_gconf_get_int (PANEL_APPLET (applet), "update_interval", NULL);
 	applet->flame_sensitivity = panel_applet_gconf_get_int (PANEL_APPLET (applet), "flame_sensitivity", NULL);
 	applet->image_size = panel_applet_gconf_get_int (PANEL_APPLET (applet), "image_size", NULL);
 
 
 	/* when gconf_get_int failes */
-	if (!applet->update_interval)
-		applet->update_interval = 50;
+	if (!applet->min_update_interval)
+		applet->min_update_interval = 50;
 	if (!applet->flame_sensitivity)
 		applet->flame_sensitivity = 1000;
 	if (!applet->image_size)
 		applet->image_size = 64;
 
 	/* check range */
-	if (applet->update_interval < 10)
-		applet->update_interval = 10;
+	if (applet->min_update_interval < 10)
+		applet->min_update_interval = 10;
 	if (applet->flame_sensitivity < 10)
 		applet->flame_sensitivity = 10;
 	if (applet->image_size < 16)
 		applet->image_size = 16;
+
+	applet->cur_update_interval = applet->min_update_interval;
+	applet->max_update_interval = 1000;
+	applet->timeout_change_time = 0;
 }
 
 static void
 cpufire_applet_save_properties (CPUFireApplet *applet)
 {
-	panel_applet_gconf_set_int (PANEL_APPLET (applet), "update_interval",   (gint) (applet->update_interval), NULL);
+	panel_applet_gconf_set_int (PANEL_APPLET (applet), "update_interval",   (gint) (applet->min_update_interval), NULL);
 	panel_applet_gconf_set_int (PANEL_APPLET (applet), "flame_sensitivity", (gint) (applet->flame_sensitivity), NULL);
 	panel_applet_gconf_set_int (PANEL_APPLET (applet), "image_size",       (gint) (applet->image_size), NULL);
 }
@@ -163,7 +174,7 @@ cpufire_applet_option_change (GtkWidget *widget, gpointer user_data)
 	w= g_object_get_data (G_OBJECT (applet->prefs_dialog), "size-spinbutton");
 	applet->image_size = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(w));
 	w= g_object_get_data (G_OBJECT (applet->prefs_dialog), "update-interval-spinbutton");
-	applet->update_interval = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(w));
+	applet->min_update_interval = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(w));
 	w= g_object_get_data (G_OBJECT (applet->prefs_dialog), "sensitivity-spinbutton");
 	applet->flame_sensitivity = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(w));
 
@@ -207,7 +218,7 @@ cpufire_applet_properties_dialog (BonoboUIComponent *uic, CPUFireApplet *applet)
 	spin_button = glade_xml_get_widget (xml, "update-interval-spinbutton");
 	gtk_object_set_data (GTK_OBJECT (applet->prefs_dialog),
 			"update-interval-spinbutton", spin_button);
-	gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_button), applet->update_interval);
+	gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_button), applet->min_update_interval);
 	g_signal_connect(GTK_OBJECT(spin_button), "value_changed",
 			GTK_SIGNAL_FUNC(cpufire_applet_option_change), applet);
 
@@ -504,19 +515,49 @@ on_darea_expose (GtkWidget *widget,
 		return FALSE;
 }
 
+static double
+get_cpu_average(CPUFireApplet* applet)
+{
+	double total_avg=0;
+	int t=applet->time_idx;
+	glibtop_cpu* cpu = &(applet->cpu[t]);
+	glibtop_get_cpu (cpu);
+
+	glibtop_cpu* c = &(applet->cpu[t]);
+	glibtop_cpu* c_ref = &(applet->cpu[t^1]);
+
+	if (c->total - c_ref->total)
+	{
+		total_avg =  (c->user - c_ref->user +
+			c->sys  - c_ref->sys +
+			c->nice - c_ref->nice) / ((double) (c->total - c_ref->total));
+	}
+
+	applet->time_idx^=1;
+
+	return total_avg;
+}
+
+
 gboolean
 burn_flame (CPUFireApplet* applet)
 {
+	struct timeval tv;
 	gint x, y;
 	gint image_width=applet->image_width;
 	gint image_height=applet->image_height;
 	int rnd_color=0;
 	int max_color=0;
-	double total_avg=0;
+	double total_avg=get_cpu_average(applet);
+	double current_time=0;
+	int current_area=0;
+	int total_area= 255.0*(image_height+2.0*applet->border)*(image_width + 2.0*applet->border);
 
 	guchar* rgbbuf = applet->rgbbuf;
 	guchar* firebuf = applet->firebuf;
-	glibtop_cpu* cpu = &(applet->cpu[applet->time_idx]);
+
+	if (gettimeofday(&tv, NULL)==0)
+		current_time=tv.tv_sec+((double)(tv.tv_usec))/1e6;
 
 	if (applet->config_change)
 	{
@@ -527,40 +568,26 @@ burn_flame (CPUFireApplet* applet)
 		g_free(applet->firebuf);
 		applet->image_width=applet->new_image_width;
 		applet->image_height=applet->new_image_height;
-		applet->firebuf= g_malloc0((applet->image_width + 2*applet->border) * (applet->image_height + 2*applet->border));
-		applet->rgbbuf= g_malloc0(applet->image_width * applet->image_height * 3);
-		gtk_widget_set_size_request (applet->drawing_area, applet->image_width, applet->image_height);
-		applet->timeout_handler_id = g_timeout_add (applet->update_interval, (GSourceFunc) burn_flame, applet);
+		applet->firebuf= g_malloc0((applet->image_width + 2*applet->border) *
+				(applet->image_height + 2*applet->border));
+
+		applet->rgbbuf= g_malloc0(applet->image_width *
+				applet->image_height * 3);
+
+		gtk_widget_set_size_request (applet->drawing_area, applet->image_width,
+				applet->image_height);
+
+		applet->timeout_handler_id = g_timeout_add (applet->min_update_interval,
+				(GSourceFunc) burn_flame, applet);
 		return FALSE;
 	}
 
-	glibtop_get_cpu (cpu);
-
-	glibtop_cpu* c = &(applet->cpu[applet->time_idx]);
-	glibtop_cpu* c_ref = &(applet->cpu[applet->ref_time_idx]);
-
-	if (c->total - c_ref->total)
-	{
-		total_avg =  (c->user - c_ref->user +
-			c->sys  - c_ref->sys +
-			c->nice - c_ref->nice) / ((double) (c->total - c_ref->total));
-	}
-
-	max_color= (int) 255.0/log(applet->flame_sensitivity)*log(1+applet->flame_sensitivity*total_avg);
+	max_color= (int) 255.0/log(applet->flame_sensitivity)
+		*log(1+applet->flame_sensitivity*total_avg);
 
 	if (max_color > 255)
 		max_color = 255;
 
-	applet->time_idx++;
-	if (applet->time_idx >= AVERAGE_NUM_TIMES)
-		applet->time_idx=0;
-
-	if (applet->time_idx == applet->ref_time_idx)
-		applet->ref_time_idx = (applet->ref_time_idx-1);
-
-	if (applet->ref_time_idx < 0)
-		applet->ref_time_idx= AVERAGE_NUM_TIMES-1;
-
 	for (x = 0; x < image_width+2*applet->border; x++)
 	{
 		for (y=0; y<applet->border; y++)
@@ -570,7 +597,10 @@ burn_flame (CPUFireApplet* applet)
 			else
 				rnd_color=0;
 
-			firebuf[(image_height+applet->border+y)*(image_width+2*applet->border)+x]=rnd_color;
+			int idx=(image_height+applet->border+y)*
+				(image_width+2*applet->border)+x;
+
+			firebuf[idx]=rnd_color;
 		}
 	}
 
@@ -578,14 +608,18 @@ burn_flame (CPUFireApplet* applet)
 	{
 		for (x = 1 ; x < image_width + 2*applet->border-1; x++)
 		{
-				
-				int val= (firebuf[y  *(image_width + 2 * applet->border)+(x)]+
-						firebuf[(y+1)*(image_width + 2 * applet->border)+(x-1)]+
-						firebuf[(y+1)*(image_width + 2 * applet->border)+(x)]+
-						firebuf[(y+1)*(image_width + 2 * applet->border)+(x+1)])/4 - applet->decay;
+
+			int val= (firebuf[y  *(image_width + 2 * applet->border)+(x)]+
+					firebuf[(y+1)*(image_width + 2 * applet->border)+(x-1)]+
+					firebuf[(y+1)*(image_width + 2 * applet->border)+(x)]+
+					firebuf[(y+1)*(image_width + 2 * applet->border)+(x+1)])/4
+				- applet->decay;
+
 			if (val < 0)
 				val=0;
+
 			applet->firebuf[y * (image_width + 2 * applet->border) + x]=val;
+			current_area+=val;
 		}
 	}
 
@@ -593,7 +627,8 @@ burn_flame (CPUFireApplet* applet)
 	{
 		for (x = 0; x < image_width; x++)
 		{
-			unsigned int col=cpufire_flame[firebuf[y * (image_width + 2 * applet->border) + (x + applet->border)]];
+			unsigned int col=cpufire_flame[firebuf[y *
+				(image_width + 2 * applet->border) + (x + applet->border)]];
 			rgbbuf[3*(y * image_width + x)+2]=   col         & 0xff;
 			rgbbuf[3*(y * image_width + x)+1]= (col >> 8)  & 0xff;
 			rgbbuf[3*(y * image_width + x)+0]= (col >> 16) & 0xff;
@@ -605,18 +640,57 @@ burn_flame (CPUFireApplet* applet)
 		if (!applet->tooltip_update--)
 		{
 			gchar *tooltip_text;
-			tooltip_text = g_strdup_printf(_("Processor:\n%4.1f%% in use"), 100.0 * total_avg);
+			tooltip_text = g_strdup_printf(_("Processor:\n%4.1f%% in use"),
+					100.0 * total_avg);
+
 			if (applet->tooltips)
-				gtk_tooltips_set_tip(applet->tooltips, GTK_WIDGET (applet), tooltip_text, tooltip_text);
+			{
+				gtk_tooltips_set_tip(applet->tooltips, GTK_WIDGET (applet),
+						tooltip_text, tooltip_text);
+			}
+
 			g_free(tooltip_text);
 			applet->tooltip_update=20;
 		}
 
-		gdk_draw_rgb_image (applet->drawing_area->window, applet->drawing_area->style->fg_gc[GTK_STATE_NORMAL],
+		gdk_draw_rgb_image (applet->drawing_area->window,
+				applet->drawing_area->style->fg_gc[GTK_STATE_NORMAL],
 				0, 0, image_width, image_height,
 				GDK_RGB_DITHER_MAX, applet->rgbbuf, image_width * 3);
 	}
 
+	/* if load is below 20% dynamically adjust timeout / full fire otherwise */
+	if ((current_time - applet->timeout_change_time>0.02) && (total_avg<powersave_load_limit ||
+			applet->cur_update_interval != applet->min_update_interval))
+	{
+
+		/* if 3d-area of fire is low (<10%) */
+		if (total_avg<powersave_load_limit && (current_area<total_area/powersave_area_limit))
+		{
+			const double c1=applet->max_update_interval;
+			const double c2=log(((double)applet->min_update_interval)/
+					applet->max_update_interval)/powersave_load_limit;
+			int new_update_interval = (int) c1*exp(c2*total_avg);
+
+			if (abs(applet->cur_update_interval - new_update_interval) > applet->min_update_interval)
+			{
+				applet->cur_update_interval=new_update_interval;
+				applet->timeout_change_time = current_time;
+				applet->timeout_handler_id = g_timeout_add (applet->cur_update_interval,
+						(GSourceFunc) burn_flame, applet);
+				return FALSE;
+			}
+		}
+		else
+		{
+			applet->cur_update_interval = applet->min_update_interval;
+			applet->timeout_change_time = current_time;
+			applet->timeout_handler_id = g_timeout_add (applet->cur_update_interval,
+					(GSourceFunc) burn_flame, applet);
+			return FALSE;
+		}
+	}
+
 	return TRUE;
 }
 
@@ -627,21 +701,21 @@ cpufire_applet_fill (CPUFireApplet* applet)
 	applet->tooltips = gtk_tooltips_new();
 	g_object_ref (applet->tooltips);
 	gtk_object_sink (GTK_OBJECT (applet->tooltips));
-	gtk_tooltips_set_tip(applet->tooltips, GTK_WIDGET (applet), "Starting...", "Starting...");
+	gtk_tooltips_set_tip(applet->tooltips, GTK_WIDGET (applet),
+			"Starting...", "Starting...");
 
-	applet->cpu= g_malloc0(sizeof(glibtop_cpu)*AVERAGE_NUM_TIMES);
+	applet->cpu= g_malloc0(2*sizeof(glibtop_cpu));
 
 	applet->rgbbuf = NULL;
 	applet->firebuf = NULL;
 	applet->drawing_area = NULL;
 
-	applet->ref_time_idx = 0;
 	applet->time_idx = 0;
 	applet->tooltip_update = 0;
 	applet->config_change = FALSE;
 
 	applet->flame_sensitivity = 1000;
-	applet->update_interval = 50;
+	applet->min_update_interval = 50;
 	applet->decay = 0;
 	applet->border = 1;
 	applet->panel_border = 2;
@@ -661,8 +735,10 @@ cpufire_applet_fill (CPUFireApplet* applet)
 			  G_CALLBACK (cpufire_applet_change_orient_cb), applet);
  	g_signal_connect (G_OBJECT (applet), "change_background",
 			  G_CALLBACK (change_background_cb), applet);
-	gtk_signal_connect (GTK_OBJECT (applet->drawing_area), "expose-event", GTK_SIGNAL_FUNC (on_darea_expose), applet);
-	applet->timeout_handler_id = g_timeout_add (applet->update_interval, (GSourceFunc) burn_flame, applet);
+	gtk_signal_connect (GTK_OBJECT (applet->drawing_area), "expose-event", 
+			GTK_SIGNAL_FUNC (on_darea_expose), applet);
+	applet->timeout_handler_id = g_timeout_add (applet->min_update_interval,
+			(GSourceFunc) burn_flame, applet);
 	gtk_tooltips_enable(applet->tooltips);
 
 	glade_gnome_init ();
@@ -689,7 +765,8 @@ cpufire_applet_factory (CPUFireApplet *applet,
 {
 	gboolean retval = FALSE;
 
-	if (iid && !strncmp (iid, "OAFIID:CPUFire_Applet", strlen("OAFIID:CPUFire_Applet")))
+	if (iid && !strncmp (iid, "OAFIID:CPUFire_Applet",
+				strlen("OAFIID:CPUFire_Applet")))
 		retval = cpufire_applet_fill (applet);
 
 	return retval;
